package Site::SQLTable;
# LICENSE
#
# Copyright (c) 2001 Steve Slaven. All rights reserved.
#
# This program is free software  and is provided "as is" without
# express or implied warranty. You can redistribute it and/or
# modify it under the same terms as Perl itself.
#

use strict;
use vars qw( @ISA %AUTOFIELDS $AUTOSETTINGS );

use PObject;
@ISA = qw( PObject );

use Site::Page;
use Site::SQLField;
use Site::Template;

$AUTOSETTINGS = 1;
%AUTOFIELDS = (
			FORMACTION => $::ENV{ SCRIPT_NAME },
			TABLE => undef,
			PKEYS => undef,
			NAMEDFIELDS => undef,
			BOUNDFIELDS => undef,
			OPTS => undef,
			DBH => undef,
			TEMPLATE => undef,
			DELETETEMPLATE => undef,
			HTMLFIELDS => undef,
			DELETEBUTTON => 'Delete',
			ADDBUTTON => 'Add',
			EDITBUTTON => 'Edit',
			DELETEYESBUTTON => 'Yes',
			DELETENOBUTTON => 'No',
			MESSAGE => undef,
			PKEYVALS => undef,
			CARRY => undef,
			WHERE => undef,
			WHEREARGS => undef,
			STARTFROM => 1,
			PERPAGE => 10,
			ROWS => undef,
			QUERY => undef,
			ORDER => undef
	);

sub OBJ_NEW
{
	my $Self = shift;

	$Self -> { NAMEDFIELDS } = {};
	$Self -> { BOUNDFIELDS } = {};
	$Self -> { PKEYS } = [];
	$Self -> { OPTS } = {};
	$Self -> { HTMLFIELDS } = [];
	$Self -> { CARRY } = {};
	$Self -> { ROWS } = [];

	return( $Self );
}

sub AddField
{
	my $Self = shift;
	my %Attr = @_;

	$Self -> { NAMEDFIELDS } -> { $Attr{ name } } =
		Site::SQLField -> new( %Attr );
	push( @{ $Self -> { HTMLFIELDS } }, $Self -> NamedFields() -> { $Attr{ name } } );

	$Self -> { NAMEDFIELDS } -> { $Attr{ name } } -> Container( $Self );

	return( $Self -> { NAMEDFIELDS } -> { $Attr{ name } } );
}

sub AddFields
{
	my $Self = shift;

	# Just add the fields
	for ( @_ )
	{
		$Self -> AddField( %{ $_ } );
	}
}

sub Display
{
	my $Self = shift;
	$Self -> _MapBoundFields();
	$Self -> _LoadTableFromOpts();

	my @ToDisplay; # <--- The array of rows to display for this run... ?

	# Loop through the query.  :)
	my @PKeys = @{ $Self -> PKeys() };
	my @Fields = $Self -> _SelFields();
	my @Values;

	# Generate the select query
	my $MagicQ;

	if( $Self -> Query() )
	{
		$MagicQ = $Self -> Query();
	}
	else
	{
		$MagicQ = 'select ' . join( ", ", @Fields ) .
			' from ' . $Self -> Table();

		# Filled in because it's used later
		$Self -> Query( $MagicQ );
	}

	if( $Self -> Where() )
	{
		$MagicQ .=
			' where ' .
			$Self -> Where();
	}

	if( $Self -> Order() )
	{
		$MagicQ .=  ' order by ' . join( ", ",
			@{ $Self -> Order() } );
	}

	my $sth = $Self -> DBH() -> prepare( $MagicQ );
	if( $Self -> WhereArgs() )
	{
		$sth -> execute( @{ $Self -> WhereArgs() } );
	}else
	{
		$sth -> execute();
	}

	my( $ThisRow, $ThisRef, $TheField );
	my $Rows = $sth -> fetchall_arrayref({});
	my( $ThisValue, $FieldsRef, $Tmp );

	for( @{ $Rows } )
	{
		# Fill fields, then pull vals into ThisRow and push
		# it onto the stack of rows	
		$ThisRef = $_;
		$ThisRow = {};

		# We have to loop through the fields because
		# of the formatting, etc
		for( keys( %{ $ThisRef } ) )
		{
			$TheField = $Self -> GetFieldBoundTo( $_ );

			$TheField -> SetValue( $ThisRef -> { $_ } );
			$ThisRow -> { 'HTMLDisplay' . $TheField -> Name() } =
				$TheField -> HTMLDisplay();
			$ThisRow -> { 'HTMLValue' . $TheField -> Name() } =
				$TheField -> Value();
		}

		# Also, for each unbound field, perform {} substitution on it
		for( @{ $Self -> HTMLFields } )
		{
			if( ! $_ -> BoundTo() )
			{
				$Tmp = $_ -> Default();
				$Tmp =~ s/\{(.*?)\}/${$ThisRef}{$1}/ge;
				$_ -> SetValue( $Tmp );
			}
		}

		# Now each field has it's value, we need to snapshot this row
		# as an HTMLFields ref for unknown looping
		$FieldsRef = [];
		for( @{ $Self -> HTMLFields() } )
		{
			push( @{ $FieldsRef },
				{
					HTMLDisplay => $_ -> HTMLDisplay()
				} );
		}
		$ThisRow -> { HTMLFields } = $FieldsRef;

		push( @{ $Self -> { ROWS } }, $ThisRow );
	}

	print $MagicQ;

	$Self -> _DisplayForm();
}

sub _WhatToDo
{
	# Look at the opts and decide what to do
	my $Self = shift;
	my $ActVal = $Self -> Opts() -> { $Self -> _Action() };
	my $ButVal = $Self -> Opts() -> { $Self -> _Button() };

	#print $ActVal . "::" . $ButVal . "<BR>";
	if( $ButVal eq $Self -> DeleteButton() )
	{
		# If they pushed delete then ask if they are sure
		return( "ASKDELETE" );
	}elsif( $ActVal eq "DELETE" )
	{
		# They said yes?
		if( $ButVal eq $Self -> DeleteYesButton() )
		{
			return( "DELETE" );
		}else
		{
			# Nope
			return( "NODELETE" )
		}
	}elsif( $ActVal eq "UPDATE" )
	{
		return( "UPDATE" );
	}elsif( $ActVal eq "ADD" )
	{
		return( "ADD" );
	}elsif( $Self -> PKeyVals() ) ### FIX ME
	{
		return( "ASKUPDATE" );
	}else
	{
		# Nothing means askadd
		return( "ASKADD" );
	}
}

sub _WhatForm
{
	# Just a logic filter for the above
	my $Self = shift;
	my $Do = $Self -> _WhatToDo();

	# This will tell what form to display and the method it should do
	if( grep( $_ eq $Do, 'ASKADD',
				'ADD',
				'UPDATE',
				'DELETE',
				'NODELETE' ) )
	{
		return( "ADD" );
	}elsif( $Do eq 'ASKUPDATE' )
	{
		return( "UPDATE" );
	}elsif( $Do eq "ASKDELETE" )
	{
		# Munge the PKEYS out of the query
		$Self -> PKeyVals( [] );
		for( @{ $Self -> PKeys } )
		{
			push( @{ $Self -> PKeyVals() }, $Self -> Opts() -> {
					$Self -> _HTMLName() . '.pkeys.' . $_ }
				);
		}
		return( "DELETE" );
	}

	return( undef );
				
}

sub _FillFields
{
	# Call proper filler based on action to take
	my $Self = shift;
	my $Do = $Self -> _WhatForm();

	# If askadd or just updated or just deleted, then fill defaults
	if( $Do eq 'ADD' )
	{
		$Self -> _FillFieldsAdd();
	}elsif( $Do eq "UPDATE" )
	{
		$Self -> _FillFieldsUpdate();
	}elsif( $Do eq "DELETE" )
	{
		$Self -> _FillFieldsUpdate();
	}
}

sub _FillFieldsUpdate
{
	# This will get values from an SQL query
	my $Self = shift();

	my @PKeys = @{ $Self -> PKeys() };
	my @Fields = $Self -> _SelFields();
	my @Values;

	# Generate the select query
	my $MagicQ = 'select ' . join( ", ", @Fields ) .
		' from ' . $Self -> Table() .
		' where ' .
		join( " AND ", map( $_ . "=?", @PKeys ) );

	my $sth = $Self -> DBH() -> prepare( $MagicQ );
	$sth -> execute( @{ $Self -> PKeyVals() } );

	my $Ref = $sth -> fetchrow_hashref();

	if( ! defined( $Ref ) )
	{
		$Self -> AddMessage( "Row now found" );
		# Should break it down a bunch here
		return( undef );
	}
	for( keys( %{ $Ref } ) )
	{
		$Self -> GetField( $_ ) -> SetValue( $Ref -> { $_ } );
	}
}

sub _FillFieldsAdd
{
	my $Self = shift;

	for( @{ $Self -> HTMLFields() } )
	{
		$_ -> SetDefault();
	}
}

sub _FillFieldsBad
{
	# This will get values from the environment
}

sub _DisplayForm
{
	my $Self = shift;

	my $Ret;

	# Performs the actual form display
	if( ! $Self -> Template() )
	{
		$Self -> Template( Site::Template -> new() );
		$Self -> Template() -> Template( $Self -> _DefaultTemplate() );
	}

	$Self -> Template() -> RegisterDefault( $Self );
	$Self -> Template() -> Display();
}

sub _DisplayDelete
{
	my $Self = shift;
	if( ! $Self -> DeleteTemplate() )
	{
		$Self -> DeleteTemplate( Site::Template -> new() );
		$Self -> DeleteTemplate() ->
			Template( $Self -> _DefaultDeleteTemplate() );
	}

	$Self -> DeleteTemplate() -> RegisterDefault( $Self );
	$Self -> DeleteTemplate() -> Display();
}

sub _DefaultTemplate
{
	# Return a default template
	return( qq{<t:var name=HTMLBegin/>
<TABLE BORDER=0>
<TR>
<t:loop name=HTMLHeaders>
<TD><t:var name=HTMLHeader/></tD>
</t:loop>
</TR>

<t:loop name=Rows>
<TR>
<t:loop name=HTMLFields>
<TD VALIGN=TOP><t:var name=HTMLDisplay/></TD>
</t:loop>
</TR>
</t:loop>

</TABLE>

<t:var name=HTMLEnd/>

} );

}

sub _DefaultDeleteTemplate
{
	return( qq{<t:var name=HTMLBegin/>
Delete it?

<TABLE BORDER=0>
<t:loop name=HTMLFields>
<TR>
<TD><t:iftrue name=HTMLLabel><B><t:var name=HTMLLabel/>:</B></t:iftrue></TD>
<TD><t:var name=HTMLDisplay/></TD>
</TR>
</t:loop>
</TABLE>

<t:var name=HTMLDeleteYes/>
<t:var name=HTMLDeleteNo/>
<t:var name=HTMLEnd/>
} );
}

sub _HTMLName
{
	my $Self = shift;
	return( join( ".", $Self -> Table(), @{ $Self -> PKeys() } ) );
}

sub _Action
{
	# The name of my action
	my $Self = shift;
	return( $Self -> _HTMLName() . ".__.action" );
}

sub HTMLBegin
{
	my $Self = shift;
	my $Action = $Self -> FormAction();
	my $Do = $Self -> _WhatForm();

	my $Ret = qq{<FORM ACTION=$Action METHOD=GET>
} . '<INPUT TYPE=HIDDEN NAME=' . $Self -> _Action() .
	' VALUE="' . $Do . '">';

	if( $Do ne "ADD" )
	{
		# Add pkey vals
		my @Vals = @{ $Self -> PKeyVals() };
		for( @{ $Self -> PKeys() } )
		{
			$Ret .= '<INPUT TYPE=HIDDEN NAME=' .
				$Self -> _HTMLName() .
				'.pkeys.' .
				$_ .
				' VALUE="' .
				Site::Page ->
					EncodeEntities( shift( @Vals ) ) .
				'">';
		}
	}

	# Spew forth the carrables!
	for ( keys( %{ $Self -> Carry() } ) )
	{
		$Ret .= '<INPUT TYPE=HIDDEN NAME=' .
			$_ .
			' VALUE="' .
			Site::Page ->
				EncodeEntities( $Self -> Carry() -> { $_ } ) .
			'">';
	}

	return( $Ret );
}


sub HTMLEnd
{
	return( qq{</FORM>} );
}

sub HTMLLabel
{
	my $Self = shift;
	my $Fname = shift;

	return( $Self -> GetField( $Fname ) -> HTMLLabel() );
}

sub HTMLField
{
	my $Self = shift;
	my $Fname = shift;

	return( $Self -> GetField( $Fname ) -> HTMLField() );
}

sub GetField
{
	my $Self = shift;
	my $Fname = shift;

	return( $Self -> NamedFields() -> { $Fname } );
}

sub HTMLAdd
{
	# The add button, if applicable
	my $Self = shift;

	if($Self -> _WhatForm() eq "ADD" )
	{
		return( '<INPUT TYPE=SUBMIT NAME=' . 
			$Self -> _Button() .
			' VALUE="' . 
			$Self -> AddButton() .
			'">' );
	}
	return( undef );
}

sub HTMLEdit
{
	# The edit button
	my $Self = shift;
	if( $Self -> _WhatForm() eq "UPDATE" )
	{
		return( '<INPUT TYPE=SUBMIT NAME=' . 
			$Self -> _Button() .
			' VALUE="' . 
			$Self -> EditButton() .
			'">' );
	}

	return( undef );
}

sub HTMLDelete
{
	# The delete button
	# Only on the update form
	my $Self = shift;

	if( $Self -> _WhatForm() eq "UPDATE" )
	{
		return( '<INPUT TYPE=SUBMIT NAME=' . 
			$Self -> _Button . 
			' VALUE="' . $Self -> DeleteButton() . '">' ); 
	}
	return( undef );
}

sub HTMLDeleteYes
{
	my $Self = shift();

	return( '<INPUT TYPE=SUBMIT NAME=' . 
			$Self -> _Button . 
			' VALUE="' . $Self -> DeleteYesButton() . '">' ); 
}

sub HTMLDeleteNo
{
	my $Self = shift();

	return( '<INPUT TYPE=SUBMIT NAME=' . 
			$Self -> _Button . 
			' VALUE="' . $Self -> DeleteNoButton() . '">' ); 
}

sub HTMLDisplay
{
	my $Self = shift;
	return( $Self -> GetField( shift() ) -> DisplayValue() );
}

sub _Button
{
	my $Self = shift;

	return( $Self -> _HTMLName() . ".__.button" );
}

sub _SelFields
{
	my $Self = shift;
	# Return the field names of the fields to mod

	return( grep( /^.+$/, map( $_ -> SQLSelect(),
				@{ $Self -> { HTMLFIELDS } } ) ) );

}

sub _ModFields
{
	my $Self = shift;
	# Return the field names of the fields to mod

	return( grep( /^.+$/, map( $_ -> SQLUpdate(),
				@{ $Self -> { HTMLFIELDS } } ) ) );

}

sub AddMessage
{
	my $Self = shift;
	$Self -> { MESSAGE } .= shift() . "\n";
}

sub _FillPKeys
{
	my $Self = shift;
	my @Arr;
	my $IsSet = 0;

	for( @{ $Self -> _HTMLPKeys() } )
	{
		push( @Arr, $Self -> Opts() -> { $_ } );
		if( defined( $Arr[ -1 ] ) )
		{
			$IsSet = 1;
		}
	}

	if( $IsSet )
	{
		$Self -> { PKEYVALS } = \@Arr;
	}
}

sub _HTMLPKeys
{
	# This will be configurable later, for now the pkeys are not fully
	# qualified
	my $Self = shift;

	return( $Self -> PKeys() );
}

sub AddCarry
{
	# Add vars to the carryo
	my $Self = shift;
	my %Carry = @_;

	for( keys( %Carry ) )
	{
		$Self -> Carry() -> { $_ } = $Carry{ $_ };
	}
}

sub PKey
{
	my $Self = shift;
	$Self -> PKeys( [ shift() ] );
}

sub pkey
{
	shift() -> PKey( @_ );
}

sub HTMLHeaders
{
	my $Self = shift;
	# Make an array for the header
	my @ret;

	for( @{ $Self -> HTMLFields() } )
	{
		push( @ret, {
				HTMLHeader =>
					$Self -> HTMLHeader( $_ -> Name() ),
			} );

	}

	return( \@ret );
}

sub HTMLHeader
{
	my $Self = shift;
	my( %NewOpts );
	my $Field = $Self -> GetField( shift() );
	my $SortOrder;

	if( ! $Field -> BoundTo() )
	{
		return( $Field -> HTMLLabel() );
	}

	# Fill out the stuff to carry
	$NewOpts{ 'query' } = Site::Page -> EscapeURI( $Self -> Query() );

	if( $Self -> Order() )
	{
		$NewOpts{ 'order' } = Site::Page -> EscapeURI( join( ",",
					$Field -> BoundTo(),
					grep( $_ ne $Field -> BoundTo(),
						@{ $Self -> Order() } ) ) );
	}

	$NewOpts{ 'where' } = Site::Page -> EscapeURI( $Self -> Where() );

	if( $Self -> WhereArgs() )
	{
		$NewOpts{ 'whereargs' } = Site::Page -> EscapeURI( join( ",", @{ $Self -> WhereArgs() } ) );
	}

	# really inefficient right now!!
	my $Cnt = 0;
	for( @{ $Self -> Order() } )
	{
		$Cnt++;
		s/\s.*$//;

		if( lc( $_ ) eq lc( $Field -> BoundTo() ) )
		{
			$SortOrder = "(" . $Cnt . ")";
		}
	}

	return( $SortOrder . '<A HREF="' .
		$Self -> FormAction() .
		'?' .
		$Self -> _Action() .
		'=SORT&' .
		join( "&", map( $Self -> _HTMLName() . ".sort." . $_ .
					'=' . $NewOpts{ $_ }, keys( %NewOpts ) ) ) .
		'">' .
		$Field -> HTMLLabel() .
		'</A>' );
}

sub _MapBoundFields
{
        my $Self = shift;

        for( @{ $Self -> HTMLFields() } )
        {
                if( $_ -> BoundTo() )
                {
                        $Self -> BoundFields() -> { $_ -> BoundTo() } = $_;
                }
        }
}

sub GetFieldBoundTo
{
        my $Self = shift;
        return( $Self -> BoundFields() -> { shift() } );
}

sub _LoadTableFromOpts
{
	# Override table options if passed in the URI
	my $Self = shift;
	my $Opts = $Self -> Opts();
	my $Sname = $Self -> _HTMLName() . ".sort.";

	if( $Opts -> { $Self -> _Action() } eq "SORT" )
	{
		$Self -> { QUERY } = $Opts -> { $Sname . "query" };
		$Self -> { WHERE } = $Opts -> { $Sname . "where" };

		$Self -> { ORDER } = [];
		for( split( /,/, $Opts -> { $Sname . "order" } ) )
		{
			push( @{ $Self -> { ORDER } }, $_ );
		}

		$Self -> { WHEREARGS } = [];
		for( split( /,/, $Opts -> { $Sname . "whereargs" } ) )
		{
			push( @{ $Self -> { WHEREARGS } }, $_ );
		}
	}
}

sub DefaultDBH
{
	shift();

	$AUTOFIELDS{ DBH } = shift();
}

sub DefaultOpts
{
	shift();

	$AUTOFIELDS{ OPTS } = shift();
}

1;
